class Error(Exception):
    pass


class InvalidBufferError(Error):
    pass


class InvalidHeaderError(Error):
    pass


class InvalidMarkerError(Error):
    pass


class InvalidRecordError(Error):
    pass


class ManifestNotFoundError(Error):
    pass


class ExtendedDataItemException(Error):
    pass


class InvalidHookIdException(Error):
    pass


class NoMoreEventsError(Error):
    pass



from dissect.etl.etl import ETL, Buffer
from dissect.etl.exceptions import (
    Error,
    ExtendedDataItemException,
    InvalidBufferError,
    InvalidHeaderError,
    InvalidHookIdException,
    InvalidMarkerError,
    InvalidRecordError,
    ManifestNotFoundError,
)

__all__ = [
    "ETL",
    "Buffer",
    "Error",
    "ExtendedDataItemException",
    "InvalidBufferError",
    "InvalidHeaderError",
    "InvalidHookIdException",
    "InvalidMarkerError",
    "InvalidRecordError",
    "ManifestNotFoundError",
]



from dissect.etl.exceptions import NoMoreEventsError
from dissect.etl.headers.event import EventHeader
from dissect.etl.headers.headers import (
    ErrorHeader,
    EventInstanceGUIDHeader,
    EventTraceHeader,
    Header,
    InvalidHeader,
    Marker,
    MessageTraceHeader,
    UnimplementedHeader,
)
from dissect.etl.headers.system import (
    CompactSystemHeader,
    PerfinfoTraceHeader,
    SystemHeader,
)
from dissect.etl.utils import c_etl_headers

HEADERS: dict[int, Header] = {
    c_etl_headers.TRACE_HEADER_TYPE_SYSTEM32: SystemHeader,
    c_etl_headers.TRACE_HEADER_TYPE_SYSTEM64: SystemHeader,
    c_etl_headers.TRACE_HEADER_TYPE_COMPACT32: CompactSystemHeader,
    c_etl_headers.TRACE_HEADER_TYPE_COMPACT64: CompactSystemHeader,
    c_etl_headers.TRACE_HEADER_TYPE_FULL_HEADER32: EventTraceHeader,
    c_etl_headers.TRACE_HEADER_TYPE_INSTANCE32: EventInstanceGUIDHeader,
    c_etl_headers.TRACE_HEADER_TYPE_TIMED: UnimplementedHeader,
    c_etl_headers.TRACE_HEADER_TYPE_ERROR: ErrorHeader,
    c_etl_headers.TRACE_HEADER_TYPE_WNODE_HEADER: UnimplementedHeader,
    c_etl_headers.TRACE_HEADER_TYPE_MESSAGE: MessageTraceHeader,
    c_etl_headers.TRACE_HEADER_TYPE_PERFINFO32: PerfinfoTraceHeader,
    c_etl_headers.TRACE_HEADER_TYPE_PERFINFO64: PerfinfoTraceHeader,
    c_etl_headers.TRACE_HEADER_TYPE_EVENT_HEADER32: EventHeader,
    c_etl_headers.TRACE_HEADER_TYPE_EVENT_HEADER64: EventHeader,
    c_etl_headers.TRACE_HEADER_TYPE_FULL_HEADER64: EventTraceHeader,
    c_etl_headers.TRACE_HEADER_TYPE_INSTANCE64: EventInstanceGUIDHeader,
}


def select_event_header(data: memoryview, etl) -> Header:
    """Select event header with marker"""
    marker_int = c_etl_headers.uint32(data[:4])

    if marker_int == 0xFFFFFFFF:
        # In this case, we reached the padding inside a buffer.
        # WHich marks it as the end of data.
        raise NoMoreEventsError("We are at the end of the events inside the buffer.")

    marker = Marker(marker_int)
    header = HEADERS.get(marker.header_type, InvalidHeader)

    return header(marker, data, etl)



from unittest.mock import Mock, patch

import pytest

from dissect.etl.headers.logfile import LogfileHeader
from dissect.etl.headers.utils import select_event_header

from .testutils import buffer_system_data


def create_header(marker):
    marker_bytes = int.to_bytes(marker, 4, "little")
    event_data = marker_bytes + b"\xAD\xDE" + buffer_system_data()[6:].tobytes()
    data = memoryview(event_data)
    return select_event_header(data, Mock())


def create_event(bit64: bool):
    marker = 0xC0010000
    if bit64:
        marker = 0xC0020000
    return create_header(marker)


def test_logfile_header_creation():
    LogfileHeader(create_event(bit64=False))


def test_logfile_wrong_hookid():
    mocked_system_header = Mock()
    mocked_system_header.hook_id = 0x1
    with pytest.raises(Exception):
        LogfileHeader(mocked_system_header)


@pytest.mark.parametrize("is_64_bit,expected_size", [(True, 0x118), (False, 0x110)])
def test_logfile_header_size(is_64_bit, expected_size):
    header = LogfileHeader(create_event(bit64=is_64_bit))
    assert header.minimal_size == expected_size
    assert len(header.header) == expected_size


def test_logfile_payload():
    header = LogfileHeader(create_event(bit64=True))
    assert header.logger_name == "PerfDiag Logger"
    assert header.log_filename == r"C:\Windows\system32\WDI\LogFiles\BootPerfDiagLogger.etl"


def test_pointer_size():
    header = LogfileHeader(create_event(bit64=True))
    assert header.pointer_size == 8


def test_cpumhz():
    header = LogfileHeader(create_event(bit64=True))
    assert header.cpu_speed_in_MHz == 0x840


@patch.object(LogfileHeader, "header", ReservedFlags=2)
def test_timscale_2(mocked_header):
    header = LogfileHeader(create_event(bit64=True))
    assert header.timestamp_scale == 1.0


@pytest.mark.parametrize(
    "perf_freq,expected_result",
    [
        (10, 1000000.0),
        (20, 500000.0),
        (40, 250000.0),
        (0xDEADBEEF, 0.0026767107138356817),
    ],
)
@patch.object(LogfileHeader, "header", ReservedFlags=1)
def test_timescale_1(mocked_header, perf_freq, expected_result):
    mocked_header.PerfFreq = perf_freq
    header = LogfileHeader(create_event(bit64=True))
    assert header.timestamp_scale == expected_result


@pytest.mark.parametrize(
    "perf_freq,expected_result",
    [(10, 1.0), (20, 0.5), (40, 0.25), (0xDEADBEEF, 2.6767107138356816e-09)],
)
@patch.object(LogfileHeader, "header", ReservedFlags=3)
def test_timescale_3(mocked_header, perf_freq, expected_result):
    mocked_header.CpuSpeedInMHz = perf_freq
    header = LogfileHeader(create_event(bit64=True))
    assert header.timestamp_scale == expected_result


def test_starttime():
    logfile_info = LogfileHeader(create_event(bit64=True))
    assert logfile_info.start_time == 0x01D70AB5FE85A7C6


def test_timestamp_base():
    system_header = create_event(bit64=True)
    logfile_info = LogfileHeader(system_header)
    assert logfile_info.timestamp_base == 132586490803526455



import pytest

from dissect.etl.exceptions import InvalidMarkerError
from dissect.etl.headers.headers import Marker


@pytest.mark.parametrize(
    "marker, expected_headertype",
    [
        (0xC0010000, 0x01),
        (0xC0020000, 0x02),
        (0xC0030000, 0x03),
        (0xC0040000, 0x04),
        (0xC0100000, 0x10),
        (0xC0110000, 0x11),
        (0xC00A0000, 0x0A),
        (0xC00B0000, 0x0B),
        (0xC0140000, 0x14),
        (0xC0150000, 0x15),
        (0xC00D0000, 0x0D),
        (0xC0120000, 0x12),
        (0xC0130000, 0x13),
        (0x90000000, 0x0F),
    ],
)
def test_marker_headertype(marker, expected_headertype):
    marker = Marker(marker)
    assert marker.header_type == expected_headertype


def test_marker_remainder():
    marker = Marker(0xC00ADEAD)
    assert marker.remainder == 0xDEAD


def test_marker_invalidstart():
    with pytest.raises(InvalidMarkerError):
        Marker(0xD0000000).header_type



from abc import abstractmethod
from datetime import datetime
from enum import IntEnum
from io import BytesIO
from typing import Any
from uuid import UUID

from dissect.cstruct.types.structure import Structure

from dissect.etl.exceptions import InvalidHeaderError, InvalidMarkerError
from dissect.etl.utils import bytes_left, c_etl_headers

BIT64_HEADERS = [
    c_etl_headers.TRACE_HEADER_TYPE_SYSTEM64,
    c_etl_headers.TRACE_HEADER_TYPE_COMPACT64,
    c_etl_headers.TRACE_HEADER_TYPE_PERFINFO64,
    c_etl_headers.TRACE_HEADER_TYPE_EVENT_HEADER64,
    c_etl_headers.TRACE_HEADER_TYPE_FULL_HEADER64,
    c_etl_headers.TRACE_HEADER_TYPE_INSTANCE64,
]


class Marker:
    """Abstracts the marker calculation for headers."""

    MARKER_MASK = 0xFF000000
    HEADER_MASK = 0x00FF0000
    REMAINDER_MASK = 0x0000FFFF

    MESSAGE_FLAGS = 0x90
    HEADER_FLAGS = 0xC0

    def __init__(self, marker: int):
        self.marker = marker

    @property
    def flags(self):
        return (self.marker & self.MARKER_MASK) >> 24

    @property
    def header_type(self):
        if self.flags == self.HEADER_FLAGS:
            return (self.marker & self.HEADER_MASK) >> 16
        elif self.flags == self.MESSAGE_FLAGS:
            return c_etl_headers.TRACE_HEADER_TYPE_MESSAGE
        else:
            raise InvalidMarkerError("Unknown flags combination")

    @property
    def remainder(self):
        return self.marker & self.REMAINDER_MASK


class Header:
    """A baseclass for the different ETL headers."""

    def __init__(self, marker: Marker, data: memoryview, etl):
        self._etl = etl
        self._marker = marker
        self._header = None
        self._payload = None
        self.data = data

        if self.size < self.minimal_size:
            raise InvalidHeaderError(
                f"Size too small for record type 0x{self._marker.header_type:02x}"
                f" from marker 0x{self.marker:08x} ({self.size} < {self.minimal_size}"
            )

    @property
    @abstractmethod
    def minimal_size(self) -> int:
        """Minimum header size."""
        pass

    @property
    @abstractmethod
    def _header_type(self) -> Structure:
        """Type of header that will get parsed."""
        pass

    @property
    def provider_id(self) -> UUID:
        """Provider that generated this event."""
        return UUID(bytes_le=self.header.ProviderId)

    @property
    def version(self) -> int:
        """The version of the event."""
        return self.header.Version

    @property
    def timestamp(self) -> datetime:
        """The timestamp of the event."""
        return self._etl.calculate_timestamp(self.time_delta)

    @property
    def time_delta(self) -> int:
        """The change in time relative to the start of the logfile."""
        return self.header.TimeDelta

    @property
    def marker(self) -> int:
        """The marker data for this event."""
        return self._marker.marker

    @property
    def is_64bit(self) -> bool:
        """A value to determine if the header is 64 or 32 bits."""
        return self._marker.header_type in BIT64_HEADERS

    @property
    def size(self) -> int:
        """The size of the event.

        In most cases this is inside the remainder field of the marker.
        """
        return self._marker.remainder

    @property
    def data_size(self) -> int:
        """The size of the payload."""
        return self.size - self.minimal_size

    @property
    def payload(self) -> memoryview:
        """Grab the payload data from the datastream."""
        if not self._payload:
            self._payload = self.data[self.minimal_size : self.size]
        return self._payload

    @property
    def header(self) -> Structure:
        """Type of header that will get parsed."""
        if not self._header:
            self._header = self._header_type(self.data[: self.minimal_size])
        return self._header

    @abstractmethod
    def additional_header_fields(self) -> dict[str, Any]:
        """Additional fields that hold interesting information.

        each header subclass defines what additional information it wants to return to a record."""
        pass

    def standard_header_fields(self) -> dict[str, Any]:
        """Some standard header information that can be retrieved for any header."""
        standard_fields = {
            "version": self.version,
            "provider_id": self.provider_id,
            "timestamp": self.timestamp,
        }
        return standard_fields

    def __repr__(self):
        standard_output = " ".join(f"{key}={value}" for key, value in self.standard_header_fields().items())
        additional_output = " ".join(f"{key}={value}" for key, value in self.additional_header_fields().items())
        return f"<{self.__class__.__name__} {standard_output} {additional_output}>"


class InvalidHeader(Header):
    """An invalid header."""

    def __init__(self, marker, data, etl):
        raise InvalidHeaderError()


class UnimplementedHeader(Header):
    """A header that isn't implemented yet."""

    pass


class EventProperty(IntEnum):
    """Defines what the message trace header can additionally find in its payload."""

    SEQUENCE = 0x0001
    GUID = 0x0002
    COMPONENT_ID = 0x0004
    TIMESTAMP = 0x0008
    PERFORMANCE_TIMESTAMP = 0x0010
    SYSTEMINFO = 0x0020
    POINTER32 = 0x0040
    POINTER64 = 0x0080


class MessageTraceHeader(Header):
    def __init__(self, marker: Marker, data: memoryview, etl):
        super().__init__(marker, data, etl)
        self.payload_offset = 0
        self.opcode = None
        self._thread_id = None
        self._process_id = None
        self._sequence_number = None
        self._provider_id = None
        self._time_delta = None
        self._parse_event_properties()

    @property
    def minimal_size(self) -> int:
        return 0x8

    @property
    def _header_type(self) -> Structure:
        return c_etl_headers.MessageHeader

    def _parse_event_properties(self):
        payload = BytesIO(self.payload)

        if self.event_property & EventProperty.SEQUENCE:
            self._sequence_number = c_etl_headers.uint32(payload)

        if self._contains_provider_id(payload):
            self._provider_id = UUID(bytes_le=payload.read(16))

        if bytes_left(payload) >= 8 and self.event_property & EventProperty.TIMESTAMP:
            self._time_delta = c_etl_headers.uint64(payload)

        if self.event_property & EventProperty.COMPONENT_ID:
            raise NotImplementedError("header.EventProperty & 4")

        if self.event_property & EventProperty.SYSTEMINFO and bytes_left(payload) >= 8:
            self._thread_id = c_etl_headers.uint32(payload)
            self._process_id = c_etl_headers.uint32(payload)

    @property
    def time_delta(self) -> int:
        return self._time_delta

    @property
    def version(self) -> int:
        return None

    @property
    def id(self) -> int:
        """The id of the message event."""
        return self.header.Id

    @property
    def event_property(self) -> int:
        """What type of payload to expect."""
        return self.header.EventProperty

    @property
    def provider_id(self) -> UUID:
        return self._provider_id

    @property
    def thread_id(self) -> int:
        return self._thread_id

    @property
    def process_id(self) -> int:
        return self._process_id

    @property
    def sequence_number(self) -> int:
        return self._sequence_number

    def _contains_provider_id(self, payload: BytesIO) -> bool:
        if not (bytes_left(payload) >= 16):
            return False
        else:
            event_property = self.event_property & (EventProperty.GUID | EventProperty.COMPONENT_ID)
            return event_property == EventProperty.GUID

    def additional_header_fields(self) -> dict[str, Any]:
        return {
            "ThreadId": self.thread_id,
            "ProcessId": self.process_id,
        }


class EventTraceHeader(Header):
    @property
    def minimal_size(self) -> int:
        return 0x30

    @property
    def _header_type(self) -> Structure:
        return c_etl_headers.EventTraceHeader

    @property
    def thread_id(self) -> int:
        """The thread id that created this event."""
        return self.header.ThreadId

    @property
    def process_id(self) -> int:
        """The process id that created this event."""
        return self.header.ProcessId

    def additional_header_fields(self) -> dict[str, Any]:
        return {
            "ThreadId": self.thread_id,
            "ProcessId": self.process_id,
        }


class EventInstanceHeader(Header):
    @property
    def minimal_size(self) -> int:
        return 0x38

    @property
    def _header_type(self) -> Structure:
        return c_etl_headers.EventInstanceHeader

    def additional_header_fields(self) -> dict[str, Any]:
        return {
            "ThreadId": self.header.ids.information.ThreadId,
            "ProcessId": self.header.ids.information.ProcessId,
        }


class EventInstanceGUIDHeader(Header):
    """A more expanded EventInstanceHeader.

    This is created from an EventInstanceHeader, but it's not quite clear which one is specifically used.
    For now, this header is default."""

    @property
    def minimal_size(self) -> int:
        return 0x48

    @property
    def _header_type(self) -> Structure:
        return c_etl_headers.EventInstanceGUIDHeader

    @property
    def thread_id(self):
        """The thread id that created this event."""
        return self.header.ThreadId

    @property
    def process_id(self):
        """The process id that created this event."""
        return self.header.ProcessId

    @property
    def parent_guid(self):
        return UUID(bytes_le=self.header.ParentGuid)

    def additional_header_fields(self) -> dict[str, Any]:
        return {
            "ThreadId": self.thread_id,
            "ProcessId": self.process_id,
            "ParentGuid": f"{self.parent_guid}",
        }


class ErrorHeader(Header):
    """When an error event was created. However, the structure of the header isn't clear."""

    @property
    def minimal_size(self) -> int:
        return 0x50

    @property
    def _header_type(self) -> Structure:
        return None

    def additional_header_fields(self) -> dict[str, Any]:
        return {}



from unittest.mock import Mock

from dissect import cstruct

from dissect.etl.utils import c_etl_definitions

RAW_BUFFER_HEADER = (
    b"\x00\x00\x10\x00\xA8\x02\x00\x00\xA8\x02\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x12\x00\x03\x00\x00\x00"
    b"\xA8\x02\x00\x00\x21\x00\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00"
)


def load_etl_definition():
    c_etl = cstruct.cstruct()
    c_etl.load(c_etl_definitions)
    return c_etl


def test_buffer_header_parsed_correctly():
    mocked_filehandle = Mock()
    mocked_filehandle.read.return_value = RAW_BUFFER_HEADER
    loaded_definition = load_etl_definition()

    header = loaded_definition.BufferHeader(mocked_filehandle)

    assert header.BufferSize == 0x100000
    assert header.SavedOffset == 0x2A8
    assert header.CurrentOffset == 0x2A8
    assert header.ReferenceCounter == 0
    assert header.TimeDelta == 0
    assert header.SequenceNumber == 0
    assert header.Defined_1 == 0
    assert header.ProcessorIndex == 0x0
    assert header.LoggerId == 0x12
    assert header.ETW_BUFFER_STATE == 0x3
    assert header.FilledBytes == 0x2A8
    assert header.BufferFlag == 0x21
    assert header.BufferType == 0x4
    assert header.unk17 == 0
    assert header.unk18 == 0
    assert header.unk19 == 0
    assert header.unk20 == 0



from enum import IntEnum
from typing import Any, OrderedDict
from uuid import UUID

from dissect.util.sid import read_sid

from dissect.etl.exceptions import ExtendedDataItemException
from dissect.etl.headers.headers import Header
from dissect.etl.utils import c_etl_headers


def read_uuid(data: bytes) -> UUID:
    uuid_data = c_etl_headers.char[16](data)
    return UUID(bytes_le=uuid_data)


def read_instance_info(data: bytes) -> OrderedDict[str, Any]:
    instance_info = c_etl_headers.EVENT_HEADER_EXT_TYPE_ITEM_INSTANCE(data)
    output_dict = instance_info._values
    parent_guid = read_uuid(output_dict.get("ParentGuid"))
    output_dict["ParentGuid"] = f"{parent_guid}"
    return output_dict


def read_stack_trace(data: bytes) -> OrderedDict[str, Any]:
    instance_info = c_etl_headers.EVENT_HEADER_EXT_TYPE_STACK_TRACE32(data)
    address_length = (len(data) - 8) // instance_info._sizes["Address"]
    instance_info._values["Address"] = c_etl_headers.uint32[address_length](data[8:])
    return instance_info._values


def read_stack_trace64(data: bytes) -> OrderedDict[str, Any]:
    instance_info = c_etl_headers.EVENT_HEADER_EXT_TYPE_STACK_TRACE64(data)
    address_length = (len(data) - 8) // instance_info._sizes["Address"]
    instance_info._values["Address"] = c_etl_headers.uint64[address_length](data[8:])
    return instance_info._values


def read_provider_traits(data: bytes) -> OrderedDict[str, Any]:
    provider_traits = c_etl_headers.EVENT_HEADER_EXT_TYPE_PROVIDER_TRAIT(data)
    output_dict = provider_traits._values
    trait_offset = sum(provider_traits._sizes.values())
    traits = []
    while trait_offset < provider_traits.TraitSize:
        trait = c_etl_headers.TRAIT(data[trait_offset:])
        traits.append(trait._values)
        trait_offset += trait.TraitSize
    output_dict.update({"Traits": traits})
    return output_dict


class EventDescriptor:
    """An representation of the Event data in a event header."""

    __slots__ = [
        "id",
        "version",
        "channel",
        "level",
        "opcode",
        "task",
        "keywords",
    ]

    def __init__(self, header):
        self.id = header.Id
        self.version = header.Version
        self.channel = header.Channel
        self.level = header.Level
        self.opcode = header.OpCode
        self.task = header.Task
        self.keywords = header.Keywords


class ExtType(IntEnum):
    RELATED_ACTIVITY_ID = 0x1
    SID = 0x2
    TS_ID = 0x3
    INSTANCE_INFO = 0x4
    STACK_TRACE32 = 0x5
    STACK_TRACE64 = 0x6
    PEBS_INDEX = 0x7
    PMC_COUNTERS = 0x8
    PSM_KEY = 0x9
    EVENT_KEY = 0xA
    EVENT_SCHEMA_TL = 0xB
    PROV_TRAITS = 0xC
    PROCESS_START_KEY = 0xD
    TYPE_MAX = 0xE
    UNKNOWN = 0x0


extended_data_item_reader = {
    ExtType.RELATED_ACTIVITY_ID: lambda x: {"Guid": f"{read_uuid(x)}"},
    ExtType.SID: lambda x: {"Sid": read_sid(x)},
    ExtType.TS_ID: lambda x: {"SessionId": c_etl_headers.uint32(x)},
    ExtType.INSTANCE_INFO: read_instance_info,
    ExtType.STACK_TRACE32: read_stack_trace,
    ExtType.STACK_TRACE64: read_stack_trace64,
    ExtType.PEBS_INDEX: lambda x: {"PebsIndex": c_etl_headers.uint32(x)},
    ExtType.PMC_COUNTERS: lambda x: {"PmcCounters": c_etl_headers.uint64[len(x) // 8](x)},
    ExtType.PSM_KEY: lambda x: {"PsmKey": c_etl_headers.uint64(x)},
    ExtType.EVENT_KEY: lambda x: {"EventKey": c_etl_headers.uint64(x)},
    ExtType.EVENT_SCHEMA_TL: lambda x: {"EventSchema": c_etl_headers.char[len(x)](x)},
    ExtType.PROV_TRAITS: read_provider_traits,
    ExtType.PROCESS_START_KEY: lambda x: {"ProcessStartKey": c_etl_headers.uint64(x)},
    ExtType.TYPE_MAX: lambda x: {"Max": c_etl_headers.char[len(x)](x)},
}


class EventHeaderExtendedDataItem:
    """Loads an extended data item from payload."""

    __slots__ = [
        "size",
        "reserved1",
        "ext_type",
        "linkage",
        "reserved2",
        "data_size",
        "data",
        "raw_data",
    ]

    def __init__(self, payload):
        header = c_etl_headers.EventHeaderExtendedDataItemHeader(payload)
        self.size = header.Size
        self.ext_type = self._extension_type(header.ExtType)
        self.reserved1 = header.Reserved1
        self.data_size = header.DataSize
        self.data = self._read_extension_type(self.ext_type, header.Data)
        self.raw_data = header.Data
        self.linkage = 0
        self.reserved2 = 0

        self.validate_header()

    def validate_header(self) -> None:
        if self.size < 8:
            raise ExtendedDataItemException("DataItem size smaller than 8 bytes.")

        if self.size % 8:
            raise ExtendedDataItemException("DataItem size not aligned with 8 bytes.")

        if self.data_size > self.size - 8:
            raise ExtendedDataItemException("Data size larger than DataItem size.")

    def _extension_type(self, item_type: int) -> ExtType:
        try:
            return ExtType(item_type)
        except ValueError:
            return ExtType.UNKNOWN

    def _read_extension_type(self, ext_type, data) -> dict:
        reader = extended_data_item_reader.get(ext_type)
        return reader(data) if reader else {}

    def __getattr__(self, name: str) -> Any:
        try:
            return object.__getattribute__(self, name)
        except AttributeError:
            pass
        return self.data.get(name)

    def __repr__(self):
        return (
            f"<EventHeaderExtendedDataItem Size={self.size} Reserved1={self.reserved1} ExtType={self.ext_type} "
            f"Linkage={self.linkage} Reserved2={self.reserved2} DataSize={self.data_size}>"
        )


class EventHeader(Header):
    @property
    def descriptor(self):
        """Event descriptor of the header."""
        return EventDescriptor(self.header)

    @property
    def header_extensions(self) -> list[EventHeaderExtendedDataItem]:
        """A list with all the extended data items for this Event."""
        return self._read_extensions()

    @property
    def minimal_size(self):
        """Minimum header size."""
        return 0x50

    @property
    def _header_type(self):
        """Type of header that will get parsed."""
        return c_etl_headers.EventHeader

    def _read_extensions(self) -> list[EventHeaderExtendedDataItem]:
        """Read header extensions from the payload"""
        count = 0
        items = []
        payload_pos = 0
        payload_size = len(self.payload)

        while True:
            if count >= 13:
                break

            if payload_pos + 8 > payload_size:
                break

            try:
                header_extension = EventHeaderExtendedDataItem(self.payload[payload_pos:])
            except (EOFError, ExtendedDataItemException):
                break

            items.append(header_extension)

            count += 1
            payload_pos += header_extension.size

        return items

    @property
    def provider_id(self):
        """Provider that generated this event."""
        return UUID(bytes=self.header.ProviderId)

    @property
    def activity_id(self):
        """The ID associated with the activity in the event.

        At least, that is my assumption."""
        return UUID(bytes_le=self.header.ActivityId)

    @property
    def opcode(self):
        """The opcode used in this event."""
        return self.header.OpCode

    @property
    def thread_id(self):
        """The thread id that created this event."""
        return self.header.ThreadId

    @property
    def process_id(self):
        """The process id that created this event."""
        return self.header.ProcessId

    def additional_header_fields(self) -> dict[str, Any]:
        basic_information = {
            "ThreadId": self.thread_id,
            "ProcessId": self.process_id,
            "ActivityId": f"{self.activity_id}",
        }
        extensions = []
        for extension in self.header_extensions:
            extension_dict = {
                "ExtType": extension.ext_type,
            }
            extension_dict.update(extension.data)
            extensions.append(extension_dict)
        basic_information.update({"Extensions": extensions})
        return basic_information



import importlib
import importlib.resources
import types
from collections import defaultdict
from pathlib import Path
from string import Formatter
from typing import BinaryIO
from uuid import UUID
from xml.etree import ElementTree

from dissect import cstruct

from dissect.etl.exceptions import ManifestNotFoundError

MODPATH = "dissect.etl.manifests"

STRUCT_FMT = """
struct {name} {{
{fields}
}};
"""

CLASS_FMT = """
from uuid import UUID
from collections import namedtuple

from dissect import cstruct
from dissect.cstruct import RawType
from dissect.cstruct import Structure

Structure._calc_offsets = lambda _: None
Keyword = namedtuple('Keyword', ['name', 'message', 'mask'])
Task = namedtuple('Task', ['name', 'message', 'value'])
Event = namedtuple('Event', ['symbol', 'value', 'version', 'opcode', 'level', 'task', 'keywords', 'template'])


class VariableType(RawType):
    def __init__(self, cstruct, type_, size):
        self._t = type_
        super().__init__(cstruct, 'VariableType', size)

    def as_64bit(self):
        raise NotImplementedError()

    def as_32bit(self):
        raise NotImplementedError()

    def _read(self, stream, context=None):
        return self._t._read(stream, context)

    def _read_array(self, stream, count, context=None):
        return self._t._read_array(stream, count, context)

    def _read_0(self, stream, context=None):
        return self._t._read_0(stream, context)

    def _write(self, stream, data):
        return self._t._write(stream, data)

    def _write_array(self, stream, data):
        return self._t._write_array(stream, data)

    def _write_0(self, stream, data):
        return self._t._write_0(stream, data)


class EtwPointer(VariableType):
    def __init__(self, cstruct):
        super().__init__(cstruct, cstruct.uint64, 8)

    def as_64bit(self):
        if self.size == 8:
            return
        self.size = 8
        self._t = self.cstruct.uint64

    def as_32bit(self):
        if self.size == 4:
            return
        self.size = 4
        self._t = self.cstruct.uint32


class UserSID_blob(VariableType):
    def __init__(self, cstruct):
        super().__init__(cstruct, cstruct.char[16], 16)

    def as_64bit(self):
        if self.size == 16:
            return
        self.size = 16
        self._t = self.cstruct.char[16]

    def as_32bit(self):
        if self.size == 8:
            return
        self.size = 8
        self._t = self.cstruct.char[8]


PROVIDER_NAME = {provider_name!r}
PROVIDER_GUID = UUID({provider_guid!r})
PROVIDER_SYMBOL = {provider_symbol!r}

c_parser = cstruct.cstruct()
c_parser.addtype("EtwPointer", EtwPointer(c_parser))
c_parser.addtype("UserSID_blob", UserSID_blob(c_parser))
c_parser.load(\"\"\"
struct SYSTEMTIME {{
    WORD    wYear;
    WORD    wMonth;
    WORD    wDayOfWeek;
    WORD    wDay;
    WORD    wHour;
    WORD    wMinute;
    WORD    wSecond;
    WORD    wMilliseconds;
}};

struct UserSID {{
    uint8 revision;
    uint8 subAuthorityCount;
    char authority[6];
    uint32 subAuthorities[subAuthorityCount];
}};

struct SID {{
    UserSID_blob    blob;
    UserSID         sid;
}};

{templates}
\"\"\")

STRINGS = {{
{strings}
}}

KEYWORDS = {{
{keywords}
}}

EVENTS = {{
{events}
}}
"""

FIELD_MAP = {
    "Boolean": "uint8 {name}",
    "GUID": "char {name}[16]",
    "IPAddrV4": "uint32 {name}",
    "IPAddrV6": "char {name}[16]",
    "Pointer": "EtwPointer {name}",
    "SInt8": "int8 {name}",
    "SInt16": "int16 {name}",
    "SInt32": "int32 {name}",
    "SInt64": "int64 {name}",
    "SizeT": "uint32 {name}",
    "Struct": "KAPUT",
    "String": "char {name}[]",
    "SYSTEMTIME": "SYSTEMTIME {name}",
    "SID": "SID {name}",
    "WString": "wchar {name}[]",
    "UnicodeString": "wchar {name}[]",
}

CACHE: dict[UUID, types.ModuleType] = {}

c_parser = cstruct.cstruct()


def lookup(guid: UUID) -> types.ModuleType:
    global CACHE

    try:
        return CACHE[guid]
    except KeyError:
        pass

    try:
        mod = importlib.import_module(f"{MODPATH}.{{{guid}}}")
        CACHE[guid] = mod
        return mod
    except ImportError:
        pass

    try:
        mod = compile_xml(guid, get_resource_string(f"manifests/xml/{{{guid}}}.xml"))
        CACHE[guid] = mod
        return mod
    except IOError:
        pass

    raise ManifestNotFoundError(guid)


def compile_file(guid: UUID, path: str) -> types.ModuleType:
    with open(path, "r") as fh:
        return compile_xml(guid, fh.read())


def compile_xml(guid: UUID, s: str) -> types.ModuleType:
    generated = generate_from_xml(s)
    # print generated
    code = compile(generated, f"<manifest {guid}>", "exec")
    module = types.ModuleType(str(guid))
    exec(code, module.__dict__)
    return module


def generate_from_file(path: str) -> str:
    with open(path, "r") as fh:
        return generate_from_xml(fh.read())


def generate_from_xml(s: str) -> str:
    e = ElementTree.fromstring(s)
    formatter = Formatter()

    strings = []
    for string in e.iter("{http://schemas.microsoft.com/win/2004/08/events}string"):
        strings.append("    '{id}': '{value}',".format(**string.attrib))

    keywords = []
    for keyword in e.iter("{http://schemas.microsoft.com/win/2004/08/events}keyword"):
        attr = keyword.attrib
        attr["message"] = attr["message"].split(".")[1][:-1]
        keywords.append("    '{name}': Keyword('{name}', STRINGS['{message}'], {mask}),".format(**attr))

    templates = []
    for template in e.iter("{http://schemas.microsoft.com/win/2004/08/events}template"):
        sname = template.attrib["tid"]

        if sname.startswith("0x"):
            sname = sname.replace("0x", "NULLx")

        fields = []
        for field in template:
            fattr = field.attrib
            fname = fattr["name"]
            ftype = fattr["inType"].split(":")[1]

            if ftype.lower() in c_parser.typedefs:
                ctype = ftype.lower()
                line = "{} {{name}}".format(ctype)
            else:
                line = FIELD_MAP[ftype]

            if "max" in fattr:
                line = "{}[{}]".format(line, fattr["max"])

            fields.append(line.format(name=fname))

        templates.append(
            STRUCT_FMT.format(
                name=sname,
                fields="\n".join(["    {};".format(field) for field in fields]),
            )
        )

    events = []
    for event in e.iter("{http://schemas.microsoft.com/win/2004/08/events}event"):
        attr = defaultdict(lambda: None)
        attr.update(event.attrib)

        if not attr["template"]:
            continue

        if attr["template"].startswith("0x"):
            attr["template"] = attr["template"].replace("0x", "NULLx")

        events.append(
            formatter.vformat(
                "    ({value}, {version}): Event('{symbol}', {value}, {version}, {opcode!r}, "
                "{level!r}, {task!r}, {keyword!r}, c_parser.{template}),",
                (),
                attr,
            )
        )

    provider = next(e.iter("{http://schemas.microsoft.com/win/2004/08/events}provider"))

    return CLASS_FMT.format(
        provider_name=provider.attrib["name"],
        provider_guid=provider.attrib["guid"],
        provider_symbol=provider.attrib.get("symbol"),
        templates="\n".join(templates),
        strings="\n".join(strings),
        keywords="\n".join(keywords),
        events="\n".join(events),
    )


def get_resource_string(path: str) -> str:
    return _get_resource_path(path).read_text()


def get_resource_stream(path: str) -> BinaryIO:
    return _get_resource_path(path).open("rb")


def _get_resource_path(path: str) -> Path:
    root = importlib.resources.files(__package__) if __package__ else Path(__file__).parent
    fpath = root.joinpath(path)

    if not fpath.exists():
        raise IOError(f"Can't find resource {path}")

    return fpath



# Sources for the logfile header:
#  - https://docs.microsoft.com/en-us/windows/win32/api/evntrace/ns-evntrace-trace_logfile_header
#  - https://www.geoffchappell.com/studies/windows/km/ntoskrnl/api/etw/tracelog/trace_logfile_header.htm
from enum import IntEnum

from dissect.cstruct import Structure, cstruct

from dissect.etl.exceptions import InvalidHookIdException
from dissect.etl.headers.system import SystemHeader

logfile_def = """
struct SYSTEMTIME {
    WORD    wYear;
    WORD    wMonth;
    WORD    wDayOfWeek;
    WORD    wDay;
    WORD    wHour;
    WORD    wMinute;
    WORD    wSecond;
    WORD    wMilliseconds;
};

struct TZInfo {
    LONG        Bias;
    wchar       StandardName[32];
    SYSTEMTIME  StandardDate;
    LONG        StandardBias;
    wchar       DaylightName[32];
    SYSTEMTIME  DaylightDate;
    LONG        DaylightBias;
};

struct TraceLogfileHeader {
    uint32  BufferSize;
    union   version_information {
        uint32 Version;
        struct VersionDetail {
            uchar MajorVersion;
            uchar MinorVersion;
            uchar SubVersion;
            uchar SubMinorVersion;
        };
    };
    uint32  ProviderVersion;
    uint32  NumberOfProcessors;
    uint64  EndTime;
    uint32  TimerResolution;
    uint32  MaximumFileSize;
    uint32  LogFileMode;
    uint32  BuffersWritten;
    union {
        char LogInstanceGuid[16];
        struct {
            uint32  StartBuffers;
            uint32  PointerSize;
            uint32  EventsLost;
            uint32  CpuSpeedInMHz;
        };
    };
    PWSTR   LoggerName;
    PWSTR   LogFileName;
    TZInfo  TimeZone;
    uint32  padding; /* The timezone info is said to be 0xB0 bytes... no clue why */
    uint64  BootTime;
    uint64  PerfFreq;
    uint64  StartTime;
    uint32  ReservedFlags;
    uint32  BufferLost;
};

struct LogFileNames {
    wchar LoggerName[];
    wchar LogFileName[];
};
"""


class ReservedFlags(IntEnum):
    PERFORMANCE_FREQ = 1
    FILETIME = 2
    CPU_FREQ = 3


class LogfileHeader:
    """The logfile header.

    It is the payload of the first event in an ETL file.

    There is also a manifest file that parses this specific header.
    However, as it is a standard event that is inside every ETL file
    (and it requires some special handling for timestamp calculation) there is a dedicated parser for it.
    """

    def __init__(self, calling_header: SystemHeader):
        if calling_header.hook_id != 0x0000:
            raise InvalidHookIdException("Called with invalid hook ID")
        self._header = None
        self.system_header = calling_header
        self.data = calling_header.payload

        # Load the logfile definition depending on the pointer type used.
        self.c_logfile = cstruct()
        self.c_logfile.load(self._create_string_pointer_type() + logfile_def)

        # At the end of the logfileheader it defines the LoggerName and the LogFileName
        # The pointers with the same name in the header do not seem to relate to these strings
        # In old versions of windows, pre 6.1, they were used to point to the strings
        # in >= 6.1 they take enumarated values of HAL_PLATFORM_TIMER_SOURCE which define some time source information.

        # The strings are defined as nullterminated strings so we read them with a cstruct from the payload.
        logfile_names = self.c_logfile.LogFileNames(self.payload)
        self.logger_name = logfile_names.LoggerName
        self.log_filename = logfile_names.LogFileName
        self.timestamp_scale = self._calculate_timestamp_scale()
        self.timestamp_base = self._calculate_timestamp_base()

    @property
    def header(self) -> Structure:
        """The parsed header of the event."""
        if not self._header:
            self._header = self.c_logfile.TraceLogfileHeader(self.data[: self.minimal_size])
        return self._header

    @property
    def payload(self) -> memoryview:
        """The payload data for the event."""
        return self.data[self.minimal_size :]

    def _create_string_pointer_type(self) -> str:
        """Return the string pointer type."""
        return f"typedef {self._get_pwstr_type()} PWSTR;"

    def _get_pwstr_type(self) -> str:
        return "uint64" if self.is_64bit else "uint32"

    @property
    def is_64bit(self) -> bool:
        # If the event header is 64-bit, so is the logheader
        return self.system_header.is_64bit

    @property
    def minimal_size(self) -> int:
        """Minimum header size."""
        return 0x118 if self.is_64bit else 0x110

    def _calculate_timestamp_scale(self) -> float:
        """Calculate the timestamp_scale variable."""
        clock_mode_dict = {
            ReservedFlags.PERFORMANCE_FREQ: 10000000 / self.perf_freq,
            ReservedFlags.FILETIME: 1.0,
            ReservedFlags.CPU_FREQ: 10.0 / self.cpu_speed_in_MHz,
        }
        return clock_mode_dict.get(self.header.ReservedFlags, 0)

    def _calculate_timestamp_base(self) -> int:
        """Calculate the timestamp_base variable.

        This is based on the start_time in the header and the delta of the system_header
        """
        return self.start_time - int(self.timestamp_scale * self.system_header.time_delta)

    @property
    def cpu_speed_in_MHz(self) -> int:
        """The CPU speed that was recorded inside the logfile header."""
        return self.header.CpuSpeedInMHz

    @property
    def perf_freq(self) -> int:
        """The performance frequency used to record the etl file."""
        return self.header.PerfFreq

    @property
    def start_time(self) -> int:
        """When the etl file started to record."""
        return self.header.StartTime

    @property
    def pointer_size(self) -> int:
        """The size of stringpointers."""
        return self.header.PointerSize

    @property
    def end_time(self) -> int:
        """The time the last event was written to the ETL file."""
        return self.header.EndTime

    @property
    def buffers_written(self) -> int:
        """The number of buffers written to the file."""
        return self.header.BuffersWritten



from abc import abstractmethod
from typing import Any
from uuid import UUID

from dissect.cstruct.types.structure import Structure

from dissect.etl.headers.headers import Header
from dissect.etl.utils import c_etl_headers, lookup_guid


class SystemSpecificHeader(Header):
    @property
    @abstractmethod
    def _minimal_size(self) -> int:
        """Minimum size for this event."""
        pass

    @property
    def size(self) -> int:
        return c_etl_headers.uint16(self.data[4:6])

    @property
    def minimal_size(self) -> int:
        """Minimum header size.

        Adds additional header bytes to the result to create a correct
        payload offset.
        """
        return self._minimal_size + self._additional_header_bytes()

    @property
    def hook_id(self) -> int:
        """An ID that identifies who created the event."""
        return (self.group << 8) | self.opcode

    @property
    def opcode(self) -> int:
        """A code used to select the type of event inside the header."""
        return self.header.OpCode

    @property
    def group(self) -> int:
        """A code used to select the provider for this header."""
        return self.header.Group

    @property
    def provider_id(self) -> UUID:
        return lookup_guid(self.header.Group, self.header.OpCode)

    def _additional_header_bytes(self) -> int:
        extra = 0
        if self.marker & 0x8000:
            extra = 8
        if self.marker & 0x0700:
            extra += 8 * ((self.marker & 0x0700) >> 8)
        return extra


class SystemHeader(SystemSpecificHeader):
    """
    Creates a System header.
    This header has the following types associated with them.
        Version
        Marker
        Size
        OpCode
        Group
        ThreadId
        ProcessId
        TimeDelta
        ProcessorTime

    """

    @property
    def _minimal_size(self) -> int:
        return 0x20

    @property
    def _header_type(self) -> Structure:
        return c_etl_headers.SystemHeader

    @property
    def process_id(self) -> int:
        """The process id that created this event."""
        return self.header.ProcessId

    @property
    def processor_time(self) -> int:
        """The time it took on the processor."""
        return self.header.ProcessorTime

    @property
    def thread_id(self) -> int:
        """The thread id that created this event."""
        return self.header.ThreadId

    def additional_header_fields(self) -> dict[str, Any]:
        return {
            "ThreadId": self.thread_id,
            "ProcessId": self.process_id,
            "ProcessorTime": self.processor_time,
        }


class CompactSystemHeader(SystemSpecificHeader):
    """
    Creates a compact system header.
    This header has the following types associated with them.
        Version
        Marker
        Size
        OpCode
        Group
        ThreadId
        ProcessId
        TimeDelta

    """

    @property
    def _header_type(self) -> Structure:
        return c_etl_headers.CompactSystemHeader

    @property
    def _minimal_size(self) -> int:
        return 0x18

    @property
    def process_id(self) -> int:
        """The process id that created this event."""
        return self.header.ProcessId

    @property
    def thread_id(self) -> int:
        """The thread id that created this event."""
        return self.header.ThreadId

    def additional_header_fields(self) -> dict[str, Any]:
        return {
            "ThreadId": self.thread_id,
            "ProcessId": self.process_id,
        }


class PerfinfoTraceHeader(SystemSpecificHeader):
    """A header that records performance events for windows."""

    @property
    def _minimal_size(self) -> int:
        return 0x10

    @property
    def _header_type(self) -> Structure:
        return c_etl_headers.PerformanceInfoHeader

    def additional_header_fields(self) -> dict[str, Any]:
        return {}



from io import BytesIO
from uuid import UUID

from dissect.cstruct import cstruct

c_global_def = """
#define EVENT_TRACE_GROUP_HEADER            0x0000
#define EVENT_TRACE_GROUP_IO                0x0100
#define EVENT_TRACE_GROUP_MEMORY            0x0200
#define EVENT_TRACE_GROUP_PROCESS           0x0300
#define EVENT_TRACE_GROUP_FILE              0x0400
#define EVENT_TRACE_GROUP_THREAD            0x0500
#define EVENT_TRACE_GROUP_TCPIP             0x0600
#define EVENT_TRACE_GROUP_JOB               0x0700
#define EVENT_TRACE_GROUP_UDPIP             0x0800
#define EVENT_TRACE_GROUP_REGISTRY          0x0900
#define EVENT_TRACE_GROUP_DBGPRINT          0x0A00
#define EVENT_TRACE_GROUP_CONFIG            0x0B00
#define EVENT_TRACE_GROUP_SPARE1            0x0C00
#define EVENT_TRACE_GROUP_WNF               0x0D00
#define EVENT_TRACE_GROUP_POOL              0x0E00
#define EVENT_TRACE_GROUP_PERFINFO          0x0F00
#define EVENT_TRACE_GROUP_HEAP              0x1000
#define EVENT_TRACE_GROUP_OBJECT            0x1100
#define EVENT_TRACE_GROUP_POWER             0x1200
#define EVENT_TRACE_GROUP_MODBOUND          0x1300
#define EVENT_TRACE_GROUP_IMAGE             0x1400
#define EVENT_TRACE_GROUP_DPC               0x1500
#define EVENT_TRACE_GROUP_CC                0x1600
#define EVENT_TRACE_GROUP_CRITSEC           0x1700
#define EVENT_TRACE_GROUP_STACKWALK         0x1800
#define EVENT_TRACE_GROUP_UMS               0x1900
#define EVENT_TRACE_GROUP_ALPC              0x1A00
#define EVENT_TRACE_GROUP_SPLITIO           0x1B00
#define EVENT_TRACE_GROUP_THREAD_POOL       0x1C00
#define EVENT_TRACE_GROUP_HYPERVISOR        0x1D00
#define EVENT_TRACE_GROUP_HYPERVISORX       0x1E00

#define EVENT_TRACE_TYPE_INFO               0x00  // Info or point event
#define EVENT_TRACE_TYPE_START              0x01  // Start event
#define EVENT_TRACE_TYPE_END                0x02  // End event
#define EVENT_TRACE_TYPE_STOP               0x02  // Stop event (WinEvent compatible)
#define EVENT_TRACE_TYPE_DC_START           0x03  // Collection start marker
#define EVENT_TRACE_TYPE_DC_END             0x04  // Collection end marker
#define EVENT_TRACE_TYPE_EXTENSION          0x05  // Extension/continuation
#define EVENT_TRACE_TYPE_REPLY              0x06  // Reply event
#define EVENT_TRACE_TYPE_DEQUEUE            0x07  // De-queue event
#define EVENT_TRACE_TYPE_RESUME             0x07  // Resume event (WinEvent compatible)
#define EVENT_TRACE_TYPE_CHECKPOINT         0x08  // Generic checkpoint event
#define EVENT_TRACE_TYPE_SUSPEND            0x08  // Suspend event (WinEvent compatible)
#define EVENT_TRACE_TYPE_WINEVT_SEND        0x09  // Send Event (WinEvent compatible)
#define EVENT_TRACE_TYPE_WINEVT_RECEIVE     0xF0  // Receive Event (WinEvent compatible)
"""
c_etl_global = cstruct()
c_etl_global.load(c_global_def)

c_etl_definitions = """
#define TRACE_HEADER_TYPE_SYSTEM32          0x01
#define TRACE_HEADER_TYPE_SYSTEM64          0x02
#define TRACE_HEADER_TYPE_COMPACT32         0x03
#define TRACE_HEADER_TYPE_COMPACT64         0x04
#define TRACE_HEADER_TYPE_FULL_HEADER32     0x0A
#define TRACE_HEADER_TYPE_INSTANCE32        0x0B
#define TRACE_HEADER_TYPE_TIMED             0x0C
#define TRACE_HEADER_TYPE_ERROR             0x0D
#define TRACE_HEADER_TYPE_WNODE_HEADER      0x0E
#define TRACE_HEADER_TYPE_MESSAGE           0x0F
#define TRACE_HEADER_TYPE_PERFINFO32        0x10
#define TRACE_HEADER_TYPE_PERFINFO64        0x11
#define TRACE_HEADER_TYPE_EVENT_HEADER32    0x12
#define TRACE_HEADER_TYPE_EVENT_HEADER64    0x13
#define TRACE_HEADER_TYPE_FULL_HEADER64     0x14
#define TRACE_HEADER_TYPE_INSTANCE64        0x15

struct SYSTEMTIME {
    WORD    wYear;
    WORD    wMonth;
    WORD    wDayOfWeek;
    WORD    wDay;
    WORD    wHour;
    WORD    wMinute;
    WORD    wSecond;
    WORD    wMilliseconds;
};

struct TimeZoneInformation {
    LONG        Bias;
    wchar       StandardName[32];
    SYSTEMTIME  StandardDate;
    LONG        StandardBias;
    wchar       DaylightName[32];
    SYSTEMTIME  DaylightDate;
    LONG        DaylightBias;
};

flag ETW_BUFFER_FLAG: uint16 {
    NORMAL           = 0x0000
    FLUSH_MARKER     = 0x0001
    EVENTS_LOST      = 0x0002
    BUFFER_LOST      = 0x0004
    RTBACKUP_CORRUPT = 0x0008
    RTBACKUP         = 0x0010
    PROC_INDEX       = 0x0020
    COMPRESSED  = 0x0040
};

enum ETW_BUFFER_TYPE: uint16 {
    GENERIC      = 0x0000
    RUNDOWN      = 0x0001
    CTX_SWAP     = 0x0002
    REFTIME      = 0x0003
    HEADER       = 0x0004
    BATCHED      = 0x0005
    EMPTY_MARKER = 0x0006
    DBG_INFO     = 0x0007
    MAXIMUM      = 0x0008
};

/* WMI_BUFFER_HEADER (latest)*/
struct BufferHeader {
    uint32           BufferSize;       /* 0x00 */
    uint32           SavedOffset;      /* 0x04 */
    uint32           CurrentOffset;    /* 0x08 */
    uint32           ReferenceCounter; /* 0x0C */
    uint64           TimeDelta;        /* 0x10 */
    int64            SequenceNumber;   /* 0x18 */
    uint64           Defined_1;        /* 0x20 */
    uint16           ProcessorIndex;   /* 0x28 ETW_BUFFER_CONTEXT */
    uint16           LoggerId;         /* 0x2A ETW_BUFFER_CONTEXT */
    uint32           ETW_BUFFER_STATE; /* 0x2C */
    uint32           FilledBytes;      /* 0x30, Filled bytes inside the buffer. */
    ETW_BUFFER_FLAG  BufferFlag;       /* 0x34 */
    ETW_BUFFER_TYPE  BufferType;       /* 0x36 */
    uint32           unk17;            /* 0x38 different for multiple iterations*/
    uint32           unk18;            /* 0x3C different for multiple iterations*/
    uint32           unk19;            /* 0x40 different for multiple iterations*/
    uint32           unk20;            /* 0x44 different for multiple iterations*/
};

/* TRACE_HEADER_TYPE_SYSTEM32, TRACE_HEADER_TYPE_SYSTEM64 */
struct SystemHeader {
    uint16  Version;            /* 0x00 */
    uint16  Marker;             /* 0x02 */
    uint16  Size;               /* 0x04 */
    uint8   OpCode;             /* 0x06 */
    uint8   Group;              /* 0x07 */
    uint32  ThreadId;           /* 0x08 */
    uint32  ProcessId;          /* 0x0c */
    uint64  TimeDelta;          /* 0x10 */
    uint64  ProcessorTime;      /* 0x18 */
};

/* TRACE_HEADER_TYPE_COMPACT32, TRACE_HEADER_TYPE_COMPACT64 */
struct CompactSystemHeader {
    uint16  Version;            /* 0x00 */
    uint16  Marker;             /* 0x02 */
    uint16  Size;               /* 0x04 */
    uint8   OpCode;             /* 0x06 */
    uint8   Group;              /* 0x07 */
    uint32  ThreadId;           /* 0x08 */
    uint32  ProcessId;          /* 0x0c */
    uint64  TimeDelta;          /* 0x10 */
};

/* TRACE_HEADER_TYPE_PERFINFO32, TRACE_HEADER_TYPE_PERFINFO64 */
struct PerformanceInfoHeader {
    uint16  Version;            /* 0x00 */
    uint16  Marker;             /* 0x02 */
    uint16  Size;               /* 0x04 */
    uint8   OpCode;             /* 0x06 */
    uint8   Group;              /* 0x07 */
    uint64  TimeDelta;          /* 0x10 */
};


/* TRACE_HEADER_TYPE_MESSAGE */
struct MessageHeader {
    uint16  Size;               /* 0x00 */
    uint16  Marker;             /* 0x02 */
    uint16  Id;                 /* 0x04 */
    uint16  EventProperty;      /* 0x06 */
};

/* TRACE_HEADER_TYPE_EVENT_HEADER32, TRACE_HEADER_TYPE_EVENT_HEADER64 */
struct EventHeader {
    uint16  Size;               /* 0x00 */
    uint16  Marker;             /* 0x02 */
    uint16  Flags;              /* 0x04 */
    uint16  EventProperty;      /* 0x06 */
    uint32  ThreadId;           /* 0x08 */
    uint32  ProcessId;          /* 0x0c */
    uint64  TimeDelta;          /* 0x10 */
    char    ProviderId[16];     /* 0x18 */
    uint16  Id;                 /* 0x28 */
    uint8   Version;            /* 0x2a */
    uint8   Channel;            /* 0x2b */
    uint8   Level;              /* 0x2c */
    uint8   OpCode;             /* 0x2d */
    uint16  Task;               /* 0x2e */
    uint64  Keywords;           /* 0x30 */
    uint64  ProcessorTime;      /* 0x38 */
    char    ActivityId[16];     /* 0x40 */
};

struct EventTraceHeader {
    uint16  Size;               /* 0x00 */
    uint16  Marker;             /* 0x02 */
    uint32  Version;            /* 0x04 */
    uint32  ThreadId;           /* 0x08 */
    uint32  ProcessId;          /* 0x0C */
    uint64  TimeDelta;          /* 0x10 */
    char    ProviderId[16];     /* 0x18 */
    uint32  KernelTime;         /* 0x28 */
    uint32  UserTime;           /* 0x2B*/
};

// An older header not used anymore
struct EventInstanceHeader {
    uint16  Size;
    uint16  Marker;
    uint32  Version;
    union {
        uint64  ThreadId;
        struct {
            uint32 ThreadId;
            uint32 ProcessId;
        } information;
    } ids;
    uint64  TimeDelta;
    uint64  RegHandle;
    uint32  InstanceId;
    uint32  ParentInstanceId;
    union {
        struct {
            uint32  KernelTime;
            uint32  UserTime;
        };
        uint64  ProcessorTime;
        struct {
            uint32  EventId;
            uint32  Flags;
        };
    };
    uint64  ParentRegHandle;
}

struct EventInstanceGUIDHeader {
    uint16  Size;
    uint16  Marker;
    uint32  Version;
    uint32  ThreadId;
    uint32  ProcessId;
    uint64  TimeDelta;
    char    ProviderId[16];
    union {
        struct {
            uint32  KernelTime;
            uint32  UserTime;
        } cpu_time;
        uint64  ProcessorTime;
        struct {
            uint32  EventId;
            uint32  Flags;
        } event_info;
    } event_metadata;
    uint32  InstanceId;
    uint32  ParentInstanceId;
    char    ParentGuid[16];
};


struct EventHeaderExtendedDataItemHeader {
    uint16  Size;
    uint16  ExtType;
    uint16  Reserved1;
    uint16  DataSize;
    char    Data[DataSize];
};

struct EVENT_HEADER_EXT_TYPE_ITEM_INSTANCE {
    uint32 InstanceId;
    uint32 ParentInstanceId;
    char ParentGuid[16];
};


struct EVENT_HEADER_EXT_TYPE_STACK_TRACE32 {
    uint64 MatchId;
    uint32 Address[];
};

struct EVENT_HEADER_EXT_TYPE_STACK_TRACE64 {
    uint64 MatchId;
    uint64 Address[];
};

struct TRAIT {
    uint16 TraitSize;           // Size of this individual trait including this field
    uint8  Type;                // ETW_PROVIDER_TRAIT_TYPE
    char   Data[TraitSize-3];   // Trait data
};

struct EVENT_HEADER_EXT_TYPE_PROVIDER_TRAIT {
    uint16 TraitSize;
    char   ProviderName[];
};


"""
c_etl_headers = cstruct()
c_etl_headers.load(c_etl_definitions)

# Flags and enumerations
BufferType = c_etl_headers.ETW_BUFFER_TYPE
BufferFlag = c_etl_headers.ETW_BUFFER_FLAG

# Known UUIDs
NullGuid = UUID("{00000000-0000-0000-0000-000000000000}")
EventTraceGuid = UUID("{68fdd900-4a3e-11d1-84f4-0000f80464e3}")
DiskIoGuid = UUID("{3d6fa8d4-fe05-11d0-9dda-00c04fd7ba7c}")
PageFaultGuid = UUID("{3d6fa8d3-fe05-11d0-9dda-00c04fd7ba7c}")
ProcessGuid = UUID("{3d6fa8d0-fe05-11d0-9dda-00c04fd7ba7c}")
FileIoGuid = UUID("{90cbdc39-4a3e-11d1-84f4-0000f80464e3}")
ThreadGuid = UUID("{3d6fa8d1-fe05-11d0-9dda-00c04fd7ba7c}")
TcpIpGuid = UUID("{9a280ac0-c8e0-11d1-84e2-00c04fb998a2}")
JobGuid = UUID("{3282fc76-feed-498e-8aa7-e70f459d430e}")
UdpIpGuid = UUID("{bf3a50c5-a9c9-4988-a005-2df0b7c80f80}")
RegistryGuid = UUID("{ae53722e-c863-11d2-8659-00c04fa321a1}")
DbgPrintGuid = UUID("{13976d09-a327-438c-950b-7f03192815c7}")
EventTraceConfigGuid = UUID("{01853a65-418f-4f36-aefc-dc0f1d2fd235}")
EventTraceSpare1 = UUID("{99134383-5248-43fc-834b-529454e75df3}")
WnfGuid = UUID("{42695762-ea50-497a-9068-5cbbb35e0b95}")
PoolGuid = UUID("{0268a8b6-74fd-4302-9dd0-6e8f1795c0cf}")
PerfinfoGuid = UUID("{ce1dbfb4-137e-4da6-87b0-3f59aa102cbc}")
HeapGuid = UUID("{222962ab-6180-4b88-a825-346b75f2a24a}")
ObjectGuid = UUID("{89497f50-effe-4440-8cf2-ce6b1cdcaca7}")
PowerGuid = UUID("{e43445e0-0903-48c3-b878-ff0fccebdd04}")
ModBoundGuid = UUID("{a9152f00-3f58-4bee-92a1-70c7d079d5dd}")
ImageLoadGuid = UUID("{2cb15d1d-5fc1-11d2-abe1-00a0c911f518}")
DpcGuid = UUID("{b2d14872-7c5b-463d-8419-ee9bf7d23e04}")
CcGuid = UUID("{7687a439-f752-45b8-b741-321aec0f8df9}")
CritSecGuid = UUID("{3ac66736-cc59-4cff-8115-8df50e39816b}")
StackWalkGuid = UUID("{def2fe46-7bd6-4b80-bd94-f57fe20d0ce3}")
UmsEventGuid = UUID("{9aec974b-5b8e-4118-9b92-3186d8002ce5}")
ALPCGuid = UUID("{45d8cccd-539f-4b72-a8b7-5c683142609a}")
SplitIoGuid = UUID("{d837ca92-12b9-44a5-ad6a-3a65b3578aa8}")
ThreadPoolGuid = UUID("{c861d0e2-a2c1-4d36-9f9c-970bab943a12}")
HypervisorTraceGuid = UUID("{7f2a405c-69b5-4bf9-a1f5-30e8f1afab5e}")
HypervisorXTraceGuid = UUID("{2ce9a149-effe-42f0-a635-a1d39e26c8f2}")

GROUP_GUID_MAP = {
    c_etl_global.EVENT_TRACE_GROUP_HEADER: EventTraceGuid,
    c_etl_global.EVENT_TRACE_GROUP_IO: DiskIoGuid,
    c_etl_global.EVENT_TRACE_GROUP_MEMORY: PageFaultGuid,
    c_etl_global.EVENT_TRACE_GROUP_PROCESS: ProcessGuid,
    c_etl_global.EVENT_TRACE_GROUP_FILE: FileIoGuid,
    c_etl_global.EVENT_TRACE_GROUP_THREAD: ThreadGuid,
    c_etl_global.EVENT_TRACE_GROUP_TCPIP: TcpIpGuid,
    c_etl_global.EVENT_TRACE_GROUP_JOB: JobGuid,
    c_etl_global.EVENT_TRACE_GROUP_UDPIP: UdpIpGuid,
    c_etl_global.EVENT_TRACE_GROUP_REGISTRY: RegistryGuid,
    c_etl_global.EVENT_TRACE_GROUP_DBGPRINT: DbgPrintGuid,
    c_etl_global.EVENT_TRACE_GROUP_CONFIG: EventTraceConfigGuid,
    c_etl_global.EVENT_TRACE_GROUP_SPARE1: EventTraceSpare1,
    c_etl_global.EVENT_TRACE_GROUP_WNF: WnfGuid,
    c_etl_global.EVENT_TRACE_GROUP_POOL: PoolGuid,
    c_etl_global.EVENT_TRACE_GROUP_PERFINFO: PerfinfoGuid,
    c_etl_global.EVENT_TRACE_GROUP_HEAP: HeapGuid,
    c_etl_global.EVENT_TRACE_GROUP_OBJECT: ObjectGuid,
    c_etl_global.EVENT_TRACE_GROUP_POWER: PowerGuid,
    c_etl_global.EVENT_TRACE_GROUP_MODBOUND: ModBoundGuid,
    c_etl_global.EVENT_TRACE_GROUP_IMAGE: ImageLoadGuid,
    c_etl_global.EVENT_TRACE_GROUP_DPC: DpcGuid,
    c_etl_global.EVENT_TRACE_GROUP_CC: CcGuid,
    c_etl_global.EVENT_TRACE_GROUP_CRITSEC: CritSecGuid,
    c_etl_global.EVENT_TRACE_GROUP_STACKWALK: StackWalkGuid,
    c_etl_global.EVENT_TRACE_GROUP_UMS: UmsEventGuid,
    c_etl_global.EVENT_TRACE_GROUP_ALPC: ALPCGuid,
    c_etl_global.EVENT_TRACE_GROUP_SPLITIO: SplitIoGuid,
    c_etl_global.EVENT_TRACE_GROUP_THREAD_POOL: ThreadPoolGuid,
    c_etl_global.EVENT_TRACE_GROUP_HYPERVISOR: HypervisorTraceGuid,
    c_etl_global.EVENT_TRACE_GROUP_HYPERVISORX: HypervisorXTraceGuid,
}


def lookup_guid(group, opcode):
    # Magic number were grabbed by reverse engineering sechost.dll
    # https://www.geoffchappell.com/studies/windows/km/ntoskrnl/api/etw/callouts/hookid.htm
    lookup = group
    if lookup == 3 and opcode == 10:
        lookup = 20

    if lookup >= 31:
        return NullGuid

    return GROUP_GUID_MAP[lookup << 8]


def bytes_left(stream: BytesIO):
    """Get number of bytes left in the buffer."""
    return stream.getbuffer().nbytes - stream.tell()



from unittest.mock import Mock

import pytest

from dissect.etl.exceptions import ExtendedDataItemException
from dissect.etl.headers.event import (
    EventHeader,
    EventHeaderExtendedDataItem,
    ExtType,
    read_provider_traits,
)
from dissect.etl.headers.headers import Marker
from dissect.etl.utils import c_etl_headers

EVENT_HEADER_EVENT = (
    b"\x12\x01\x13\xC0\x01\x00\x00\x00\xE8\x02\x00\x00\xE4\x02\x00\x00"
    b"\x2B\x15\x41\x0B\x00\x00\x00\x00\xA7\xCE\xDD\xB8\x20\xB5\x09\x49"
    b"\xBC\xEB\xE0\x17\x0C\x9F\x0E\x99\x00\x00\x00\x0B\x05\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x20\x00\x00\x11\x00\x00\x00\x02\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x48\x00\x0C\x00\x01\x00\x3D\x00\x3D\x00\x4D\x69\x63\x72\x6F\x73"
    b"\x6F\x66\x74\x2E\x57\x69\x6E\x64\x6F\x77\x73\x2E\x53\x65\x72\x76"
    b"\x69\x63\x65\x43\x6F\x6E\x74\x72\x6F\x6C\x4D\x61\x6E\x61\x67\x65"
    b"\x72\x00\x13\x00\x01\x1A\x73\x50\x4F\xCF\x89\x82\x47\xB3\xE0\xDC"
    b"\xE8\xC9\x04\x76\xBA\x00\x00\x00\x38\x00\x0B\x00\x00\x00\x2C\x00"
    b"\x2C\x00\x00\x53\x65\x72\x76\x69\x63\x65\x48\x6F\x73\x74\x53\x74"
    b"\x61\x72\x74\x65\x64\x00\x53\x65\x72\x76\x69\x63\x65\x48\x6F\x73"
    b"\x74\x4E\x61\x6D\x65\x00\x01\x50\x49\x44\x00\x08\x00\x00\x00\x00"
    b"\x43\x00\x3A\x00\x5C\x00\x57\x00\x69\x00\x6E\x00\x64\x00\x6F\x00"
    b"\x77\x00\x73\x00\x5C\x00\x73\x00\x79\x00\x73\x00\x74\x00\x65\x00"
    b"\x6D\x00\x33\x00\x32\x00\x5C\x00\x73\x00\x70\x00\x70\x00\x73\x00"
    b"\x76\x00\x63\x00\x2E\x00\x65\x00\x78\x00\x65\x00\x00\x00\xF4\x01"
    b"\x00\x00\x00\x00\x00\x00\x00\x00"
)


def create_header():
    header_data = memoryview(EVENT_HEADER_EVENT)
    marker = Marker(0xC000A0112)
    return EventHeader(marker, header_data, Mock())


def test_event_header():
    event = create_header()
    header = event.header

    assert header.Size == 0x112
    assert header.Marker == 0xC013
    assert header.Flags == 0x1
    assert header.EventProperty == 0x0
    assert header.ThreadId == 0x2E8
    assert header.ProcessId == 0x2E4
    assert header.TimeDelta == 0x0B41152B
    assert header.ProviderId == b"\xA7\xCE\xDD\xB8\x20\xB5\x09\x49\xBC\xEB\xE0\x17\x0C\x9F\x0E\x99"
    assert header.Id == 0x0000
    assert header.Version == 0x00
    assert header.Channel == 0x0B
    assert header.Level == 0x05
    assert header.OpCode == 0x0
    assert header.Task == 0x0
    assert header.Keywords == 0x200000000000
    assert header.ProcessorTime == 0x0200000011
    assert header.ActivityId == bytes([0x00] * 16)


def test_event_payload():
    event = create_header()
    assert len(event.payload) == 0x112 - event.minimal_size


@pytest.mark.parametrize(
    "size, data_size, exception_match",
    [
        (0, 0, r".*smaller than 8 bytes."),
        (9, 0, r".*not aligned with 8 bytes."),
        (16, 20, r".*larger than DataItem size."),
    ],
)
def test_event_small_dataitem(size, data_size, exception_match):
    event_data = create_extended_item_header(Size=size, DataSize=data_size)
    with pytest.raises(ExtendedDataItemException, match=exception_match):
        EventHeaderExtendedDataItem(event_data)


def create_extended_item_header(**kwargs):
    if "Data" in kwargs and "DataSize" not in kwargs:
        data_size = len(kwargs["Data"])
        data = kwargs["Data"]
        kwargs.pop("Data")
    elif "DataSize" in kwargs and "Data" not in kwargs:
        data_size = kwargs["DataSize"]
        data = b"\x00" * data_size
        kwargs.pop("DataSize")

    elif "DataSize" not in kwargs and "Data" not in kwargs:
        data_size = 8
        data = b"\x00" * data_size
    else:
        data_size = kwargs["DataSize"]
        data = kwargs["Data"]
        kwargs.pop("DataSize")
        kwargs.pop("Data")

    if "Size" in kwargs:
        size = kwargs["Size"]
        kwargs.pop("Size")
    else:
        size = data_size + 8
    return c_etl_headers.EventHeaderExtendedDataItemHeader(Size=size, DataSize=data_size, Data=data, **kwargs).dumps()


def test_event_header_extension():
    EventHeaderExtendedDataItem(memoryview(create_extended_item_header()))


@pytest.mark.parametrize(
    "item_type, expected_value",
    [
        (0x1, "RELATED_ACTIVITY_ID"),
        (0x2, "SID"),
        (0x3, "TS_ID"),
        (0x4, "INSTANCE_INFO"),
        (0x5, "STACK_TRACE32"),
        (0x6, "STACK_TRACE64"),
        (0x7, "PEBS_INDEX"),
        (0x8, "PMC_COUNTERS"),
        (0x9, "PSM_KEY"),
        (0xA, "EVENT_KEY"),
        (0xB, "EVENT_SCHEMA_TL"),
        (0xC, "PROV_TRAITS"),
        (0xD, "PROCESS_START_KEY"),
        (0xE, "TYPE_MAX"),
        (0x0, "UNKNOWN"),
        (0xF, "UNKNOWN"),
    ],
)
def test_extension_ext_type(item_type, expected_value):
    header = create_extended_item_header()

    data = EventHeaderExtendedDataItem(header)
    output = data._extension_type(item_type).name

    assert output == expected_value


@pytest.mark.parametrize(
    "item_type, expected_value",
    [
        (ExtType.RELATED_ACTIVITY_ID, {"Guid": "00000000-0000-0000-0000-000000000000"}),
        (ExtType.SID, {"Sid": "S-0-0"}),
        (ExtType.TS_ID, {"SessionId": 0}),
        (
            ExtType.INSTANCE_INFO,
            {"InstanceId": 0, "ParentInstanceId": 0, "ParentGuid": "00000000-0000-0000-0000-000000000000"},
        ),
        (ExtType.STACK_TRACE32, {"MatchId": 0, "Address": [0, 0, 0, 0]}),
        (ExtType.STACK_TRACE64, {"MatchId": 0, "Address": [0, 0]}),
        (ExtType.PEBS_INDEX, {"PebsIndex": 0}),
        (ExtType.PMC_COUNTERS, {"PmcCounters": [0, 0, 0]}),
        (ExtType.PSM_KEY, {"PsmKey": 0}),
        (ExtType.EVENT_KEY, {"EventKey": 0}),
        (ExtType.EVENT_SCHEMA_TL, {"EventSchema": b"\x00" * 24}),
        (ExtType.PROV_TRAITS, {"TraitSize": 0, "ProviderName": b"", "Traits": []}),
        (ExtType.PROCESS_START_KEY, {"ProcessStartKey": 0}),
        (ExtType.TYPE_MAX, {"Max": b"\x00" * 24}),
        (ExtType.UNKNOWN, {}),
    ],
)
def test_extension_output_data(item_type, expected_value):
    header = create_extended_item_header()
    data = EventHeaderExtendedDataItem(header)
    input_data = b"\x00" * 24
    assert data._read_extension_type(item_type, input_data) == expected_value


def test_extension_key_in_data():
    header = create_extended_item_header(ExtType=ExtType.RELATED_ACTIVITY_ID, Data=b"\x00" * 16)
    data = EventHeaderExtendedDataItem(header)
    assert data.Guid == "00000000-0000-0000-0000-000000000000"


def test_extension_key_not_in_data():
    header = create_extended_item_header(ExtType=ExtType.UNKNOWN, Data=b"\x00" * 16)
    data = EventHeaderExtendedDataItem(header)
    assert data.Guid is None


def test_extension_provider_trait():
    trait_data = (
        b"=\x00Microsoft.Windows.ServiceControlManager\x00\x13\x00\x01\x1asPO\xcf\x89\x82G\xb3\xe0\xdc\xe8\xc9\x04v\xba"
    )
    traits = read_provider_traits(trait_data)
    assert traits["TraitSize"] == 0x3D
    assert traits["ProviderName"] == b"Microsoft.Windows.ServiceControlManager"
    assert traits["Traits"][0]["TraitSize"] == 0x13
    assert traits["Traits"][0]["Type"] == 0x1
    assert traits["Traits"][0]["Data"] == b"\x1asPO\xcf\x89\x82G\xb3\xe0\xdc\xe8\xc9\x04v\xba"



from unittest.mock import Mock

import pytest

from dissect.etl.exceptions import InvalidHeaderError
from dissect.etl.headers.utils import select_event_header


def create_header(marker: int, min_size: int = 0, other_bytes=b"\xAD\xDE"):
    marker_bytes = int.to_bytes(marker, 4, "little")
    amount_of_padding = min_size - len(marker_bytes) - len(other_bytes)
    padding_bytes = b"\x00" * amount_of_padding

    data = memoryview(marker_bytes + other_bytes + padding_bytes)
    header = select_event_header(data, Mock())

    try:
        additional_padding = b"\x00" * (len(header._header_type) - amount_of_padding)
    except TypeError:
        additional_padding = b""

    header.data = memoryview(marker_bytes + other_bytes + padding_bytes + additional_padding)
    return header


@pytest.mark.parametrize(
    "marker, expected_size",
    [
        (0xC0010000, 0x20),
        (0xC0020000, 0x20),
        (0xC0030000, 0x18),
        (0xC0040000, 0x18),
        (0xC00AADDE, 0x30),
        (0xC00BADDE, 0x48),
        (0xC00DADDE, 0x50),
        (0xC0100000, 0x10),
        (0xC0110000, 0x10),
        (0xC012ADDE, 0x50),
        (0xC013ADDE, 0x50),
        (0xC014ADDE, 0x30),
        (0xC015ADDE, 0x48),
    ],
)
def test_header_size(marker, expected_size):
    header = create_header(marker=marker, min_size=expected_size)
    assert header.minimal_size == expected_size


def test_message_headersize():
    header = create_header(marker=0x9000ADDE, other_bytes=b"\x00\x00\x00\x00")
    assert header.minimal_size == 0x8


def test_header_size_events():
    header = create_header(0xC00ADEAD)
    assert header.size == 0xDEAD


@pytest.mark.parametrize(
    "marker",
    [
        0xC0010000,
        0xC0020000,
        0xC0030000,
        0xC0040000,
        0xC0100000,
        0xC0110000,
    ],
)
def test_header_kernel_event_size(marker):
    header = create_header(marker)
    assert header.size == 0xDEAD


def test_invalid_header():
    with pytest.raises(InvalidHeaderError):
        create_header(0xC0FF0000)


def test_invalid_size():
    with pytest.raises(InvalidHeaderError):
        create_header(0xC00A0000)


@pytest.mark.parametrize(
    "marker",
    [
        0xC0020000,
        0xC0040000,
        0xC0110000,
        0xC013DEAD,
        0xC014DEAD,
        0xC015DEAD,
    ],
)
def test_is_64bit(marker):
    header = create_header(marker=marker, min_size=80)
    assert header.is_64bit


@pytest.mark.parametrize(
    "marker",
    [
        0xC0010000,
        0xC0030000,
        0xC0100000,
        0xC00ADEAD,
        0xC00BDEAD,
        0xC0100000,
        0xC012DEAD,
        0xC00DDEAD,
    ],
)
def test_is_32bit(marker):
    header = create_header(marker=marker, min_size=80)
    assert not header.is_64bit


@pytest.mark.parametrize(
    "marker, expected_size",
    [
        (0xC0018000, 8),
        (0xC0010700, 8 * 7),
        (0xC0018700, 8 * 8),
        (0xC0010600, 8 * 6),
        (0xC0010000, 0),
    ],
)
def test_additional_header_size(marker, expected_size):
    sys_header = create_header(marker)
    assert sys_header._additional_header_bytes() == expected_size
    assert sys_header.minimal_size == 0x20 + expected_size


@pytest.mark.parametrize(
    "marker, expected_fields",
    [
        (0xC0010000, {"ThreadId": 0, "ProcessId": 0, "ProcessorTime": 0}),
        (0xC0020000, {"ThreadId": 0, "ProcessId": 0, "ProcessorTime": 0}),
        (0xC0030000, {"ThreadId": 0, "ProcessId": 0}),
        (0xC0040000, {"ThreadId": 0, "ProcessId": 0}),
        (0xC00AADDE, {"ThreadId": 0, "ProcessId": 0}),
        (0xC00BADDE, {"ThreadId": 0, "ProcessId": 0, "ParentGuid": "00000000-0000-0000-0000-000000000000"}),
        (0xC00DADDE, {}),
        (0xC0100000, {}),
        (0xC0110000, {}),
        (
            0xC012ADDE,
            {"ThreadId": 0, "ProcessId": 0, "ActivityId": "00000000-0000-0000-0000-000000000000", "Extensions": []},
        ),
        (
            0xC013ADDE,
            {"ThreadId": 0, "ProcessId": 0, "ActivityId": "00000000-0000-0000-0000-000000000000", "Extensions": []},
        ),
        (0xC014ADDE, {"ThreadId": 0, "ProcessId": 0}),
        (0xC015ADDE, {"ThreadId": 0, "ProcessId": 0, "ParentGuid": "00000000-0000-0000-0000-000000000000"}),
    ],
)
def test_additional_fields(marker, expected_fields):
    header = create_header(marker)
    assert header.additional_header_fields() == expected_fields



from unittest.mock import MagicMock, patch

import pytest
from dissect import cstruct

import dissect.etl.etl as etl
from dissect.etl.etl import Buffer
from dissect.etl.exceptions import InvalidBufferError
from dissect.etl.utils import c_etl_definitions

from .test_wmi_buffer_header import RAW_BUFFER_HEADER as RAW_BUFFER_HEADER


def load_etl_definition():
    c_etl = cstruct.cstruct()
    c_etl.load(c_etl_definitions)
    return c_etl


def setup_mocked_etlfile(start_offset=0):
    mocked_etlfile = MagicMock()
    mocked_etlfile.fh.read.return_value = RAW_BUFFER_HEADER
    mocked_etlfile.fh.len.return_value = len(RAW_BUFFER_HEADER)
    mocked_etlfile.fh.tell.return_value = start_offset + 0x48
    return mocked_etlfile


def setup_buffer_with_offsets(data_offset, start_offset=0):
    mocked_etl = setup_mocked_etlfile()
    etl.c_etl = MagicMock()
    etl.c_etl.BufferHeader.return_value.FilledBytes = data_offset
    etl.c_etl.BufferHeader.return_value.BufferSize = data_offset
    etl.c_etl.BufferHeader.__len__.return_value = 0x48
    return Buffer(mocked_etl, start_offset)


def setup_buffer_iterator(mocked_read_record, data_offset):
    mocked_read_record.next_offset = 10
    buffer = setup_buffer_with_offsets(data_offset=data_offset)
    buffer_iterator = iter(buffer)
    return buffer_iterator


def test_buffer_data_len():
    mocked_etl = setup_mocked_etlfile()
    mocked_etl.c_etl = load_etl_definition()

    buffer = Buffer(mocked_etl, 0)
    buffer.data

    mocked_etl.fh.read.assert_called_with(buffer.filled_bytes - buffer.data_offset)


@patch.object(Buffer, "read_record")
def test_buffer_iteration_succeed(mocked_read_record):
    buffer_iterator = setup_buffer_iterator(mocked_read_record, data_offset=15)
    assert next(buffer_iterator) == mocked_read_record.return_value


@pytest.mark.parametrize("header_offset,data_offset", [(0, 10), (0, 0)])
@patch.object(Buffer, "read_record", side_effect=[EOFError])
def test_buffer_iterations(mocked_read_record, header_offset, data_offset):
    buffer_iterator = setup_buffer_iterator(mocked_read_record, data_offset)
    with pytest.raises(StopIteration):
        next(buffer_iterator)


def test_negative_read_offset():
    buffer = setup_buffer_with_offsets(start_offset=10, data_offset=5)
    with pytest.raises(InvalidBufferError):
        buffer.data


def test_different_buffer_sizes():
    buffer1 = Buffer(MagicMock(), 0)
    buffer1._header = MagicMock()



from unittest.mock import Mock

import pytest

from dissect.etl.exceptions import InvalidMarkerError
from dissect.etl.headers.utils import select_event_header

SYSTEM_EVENT = (
    b"\x02\x00\x02\xC0\x44\x00\x05\x00\xB4\x00\x00\x00\x04\x00\x00\x00"
    b"\x8F\x8C\x29\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00"
    b"\x07\x03\x00\x00\x04\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x50\x00\x00\x00\x00\x00\x00\x00"
)


def create_event(EVENT: bytearray):
    header_data = memoryview(EVENT)
    return select_event_header(header_data, Mock())


# Parse the system header as I imagined it should
def test_system_header():
    event = create_event(SYSTEM_EVENT)
    header = event.header
    assert header.Version == 0x0002
    assert header.Marker == 0xC002
    assert header.Size == 0x44
    assert header.OpCode == 0x5
    assert header.Group == 0x0
    assert header.ThreadId == 0xB4
    assert header.ProcessId == 0x4
    assert header.TimeDelta == 0x298C8F
    assert header.ProcessorTime == 0x1


def test_parse_failed():
    with pytest.raises(InvalidMarkerError):
        create_event(SYSTEM_EVENT[1:])



# Resources:
#   https://www.geoffchappell.com/studies/windows/km/ntoskrnl/api/etw/callouts/hookid.htm
#   https://www.geoffchappell.com/studies/windows/km/ntoskrnl/api/etw // For most of the header definitions
#   https://docs.microsoft.com/en-us/windows/win32/etw/wnode-header // for timestamp calculation
#   https://docs.rs/winapi-ui-automation/0.3.9/winapi_ui_automation/um/evntcons/index.html
#   Windows Development Kit
#
# TMF parser:
#   https://searchcode.com/file/116783567/Tools/ETW/traceEvent/WPPTraceEventParser.cs
#
# Maybe interesting files to look at:
#   AppData\\Local\\Microsoft\\Windows\\Explorer\\ExplorerStartupLog.etl
#   Windows\\Performance\\WinSAT\\DataStore
#   Windows\\System32\\NDF\\*.etl
#   Windows\\System32\\LogFiles\\WMI
#   Windows\\System32\\SleepStudy
#   Windows\\Panther\\*.etl
#   Windows\\System32\\WDI\\LogFiles
from __future__ import annotations

import io
from datetime import datetime
from typing import Any, Iterable, Optional
from uuid import UUID

from dissect.cstruct import Instance, cstruct
from dissect.util.compression.lzxpress import decompress as xpress_decompress
from dissect.util.sid import read_sid
from dissect.util.ts import wintimestamp

from dissect.etl import manifest
from dissect.etl.exceptions import (
    InvalidBufferError,
    InvalidHeaderError,
    ManifestNotFoundError,
    NoMoreEventsError,
)
from dissect.etl.headers.headers import Header
from dissect.etl.headers.logfile import LogfileHeader
from dissect.etl.headers.utils import select_event_header
from dissect.etl.utils import BufferFlag, c_etl_definitions

c_etl = cstruct()
c_etl.load(c_etl_definitions)


class ETL:
    """The main interface when controlling an ETL file."""

    def __init__(self, fh):
        self.fh = fh

        # Load the first buffer inside the file,
        # The first event of the buffer contains a logfile header
        first_buffer = Buffer(self, 0)
        self.buffer_header = first_buffer.header
        trace_header = first_buffer.read_record(0).header
        self.logfile_header = LogfileHeader(trace_header)

        # Lookup PointerSize
        self.pointer_size = self.logfile_header.pointer_size

        if self.pointer_size not in (4, 8):
            raise InvalidHeaderError(f"invalid pointer size: {self.pointer_size}")

        self.is_64bit = self.logfile_header.is_64bit

        self.start_time = self.logfile_header.start_time
        self.buffer_size = self.buffer_header.BufferSize
        self.start = wintimestamp(self.logfile_header.start_time)
        self.end = wintimestamp(self.logfile_header.end_time)

        self._buffer_cache = {0: first_buffer}

    def buffer(self, index: int) -> Buffer:
        """Reads a specific buffer into memory."""

        if index < 0 or index >= self.logfile_header.buffers_written:
            raise IndexError("buffer index out of range")

        try:
            buf = self._buffer_cache[index]
        except KeyError:
            prev_buf = self.buffer(index - 1)
            buf = Buffer(self, prev_buf.next_buffer)
            self._buffer_cache[index] = buf

        return buf

    def buffers(self) -> Iterable[Buffer]:
        for i in range(self.logfile_header.buffers_written):
            yield self.buffer(i)

    def __iter__(self) -> Iterable[Event]:
        for buffer in self.buffers():
            for event in buffer:
                yield event

    def calculate_timestamp(self, time_delta: int) -> datetime:
        return wintimestamp(self.get_filetime_for_event(time_delta))

    def get_filetime_for_event(self, time_delta: int) -> int:
        return self.logfile_header.timestamp_base + int(self._calculate_timescale_for_event(time_delta))

    def _calculate_timescale_for_event(self, time_delta: int) -> float:
        return self.logfile_header.timestamp_scale * time_delta


class Buffer:
    def __init__(self, etl, offset):
        self.fh = etl.fh
        self.etl = etl
        self.offset = offset

        self._header = None
        self._data = None

    @property
    def header(self):
        if not self._header:
            self.fh.seek(self.offset)
            self._header = c_etl.BufferHeader(self.fh)

        return self._header

    @property
    def size(self) -> int:
        return self.header.BufferSize

    @property
    def data(self) -> memoryview:
        if not self._data:
            data_len = min(self.size, self.filled_bytes) - len(c_etl.BufferHeader)
            if data_len < 0:
                raise InvalidBufferError("Invalid data length")
            self.fh.seek(self.data_offset)

            tmp_data = self.fh.read(data_len)

            if self.header.BufferFlag & BufferFlag.COMPRESSED:
                tmp_data = xpress_decompress(tmp_data)

            self._data = memoryview(tmp_data)
        return self._data

    @property
    def data_offset(self) -> int:
        return self.offset + len(c_etl.BufferHeader)

    @property
    def filled_bytes(self) -> int:
        return self.header.FilledBytes

    @property
    def next_buffer(self) -> int:
        return self.offset + self.size

    def __iter__(self) -> Iterable[EventRecord]:
        offset = 0
        while offset < self.filled_bytes:
            try:
                event = self.read_record(offset)
                offset += event.aligned_size
                yield event
            except (EOFError, NoMoreEventsError):
                break

    def read_record(self, offset):
        """Parse a record from a given offset inside a buffer."""

        event_record = EventRecord()
        data = self.data[offset:]
        header = select_event_header(
            data=data,
            etl=self.etl,
        )

        event_record._header = header

        return event_record

    def open(self):
        return io.BytesIO(self.data)


class EventRecord:
    __slots__ = (
        "_header",
        "_event",
    )

    def __init__(self):
        self._event = None
        self._header = None

    @property
    def header(self) -> Header:
        """A header of the type Header"""
        return self._header

    @property
    def size(self) -> int:
        """Size of the whole record."""
        return self.header.size

    @property
    def event(self) -> Event:
        """Parse payload inside the event header."""
        if not self._event:
            self._event = parse_payload(self._header)

        return self._event

    @property
    def aligned_size(self):
        return (self.size + 7) & 0xFFFFFFF8

    def __repr__(self):
        return "<EventRecord>"


class Event:
    __slots__ = [
        "_record",
        "_manifest",
        "_struct",
        "_event",
        "_header",
    ]

    def __init__(self, header: Header, event_manifest):
        self._header = header
        self._manifest = event_manifest
        self._event = None

        # Not every header has an opcode and version, so check it like this.
        opcode = getattr(header, "opcode", None)
        version = getattr(header, "version", None)
        key = (opcode, version)

        if event_manifest:
            try:
                self._event = event_manifest.EVENTS[key]
                self._struct = self._event.template(header.payload)
            except KeyError:
                self._struct = None
        else:
            self._struct = None

    def __getattr__(self, attribute: str):
        try:
            return getattr(self._struct, attribute)
        except AttributeError:
            return object.__getattribute__(self, attribute)

    def provider_name(self) -> Optional[str]:
        """Returns the manifest provider name."""
        return self._manifest.PROVIDER_NAME if self._manifest else None

    def ts(self) -> datetime:
        """Returns the event timestamp."""
        return self._header.timestamp

    def provider_id(self) -> UUID:
        """Returns the GUID of the provider from the header."""
        return self._header.provider_id

    def symbol(self):
        return self._event.symbol if self._event else None

    def event_values(self) -> dict[str, Any]:
        """Create an items view that holds event and header data.

        The header data is additional information provided from a specific header.
        The event data is from a specific manifest file if it exists.
        """
        event_values = self._header.additional_header_fields()
        struct_events = self._struct._values if self._struct else {}

        # Pretty print Instance values.
        update_event = {}
        instance_values = ((key, value) for key, value in struct_events.items() if isinstance(value, Instance))
        for key, value in instance_values:
            if hasattr(value, "sid"):
                update_event[key] = read_sid(value.sid.dumps())

        event_values.update(struct_events)
        event_values.update(update_event)
        return event_values

    def __repr__(self):
        symbol = self._event.symbol if self._event else None
        return f"{self._header} <{symbol} {self._struct!r}>"


def parse_payload(header: Header):
    """Parse the event payload using the appropriate manifest, if available."""

    try:
        mf = manifest.lookup(header.provider_id)
        if header.is_64bit:
            mf.c_parser.EtwPointer.as_64bit()
            mf.c_parser.UserSID_blob.as_64bit()
        else:
            mf.c_parser.EtwPointer.as_32bit()
            mf.c_parser.UserSID_blob.as_32bit()
    except (ManifestNotFoundError, AttributeError):
        mf = None

    return Event(header, mf)
